#define CASIO_PV /* Activated to compile on Casio Pocket Viewer SDK */
/******************************************************************************
  Include di sistema
******************************************************************************/
#ifdef CASIO_PV
#include <stdrom.h>
#include "dummy.h" /* Used to compile with math functions */
#include <ctype.h>
#include	"define.h"
#include	"libc.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#endif

/******************************************************************************
  Include applicazione
******************************************************************************/
#include "Parser.h"

/******************************************************************************
  Variabili globali
******************************************************************************/
/* Tabella di parsing */
byte parsetable[tMaxType][tMaxType] = {
    /*                ------------- INPUT -------------- */
    /*                +   *   ^   M   f   ,   (   )   $  */
    /*  STACK         --  --  --  --  --  --  --  --  -- */
    /* tPlusMin */  { R,  S,  S,  S,  S,  R,  S,  R,  R },
    /* tMultDiv */  { R,  R,  S,  S,  S,  R,  S,  R,  R },
    /* tExpon   */  { R,  R,  S,  S,  S,  R,  S,  R,  R },
    /* tUminus  */  { R,  R,  R,  S,  S,  R,  S,  R,  R },
    /* tFunct   */  { R,  R,  S,  E4, S,  E4, S,  R,  R },
    /* tComma   */  { R,  R,  R,  R,  R,  E4, R,  R,  E4},
    /* tLPar    */  { S,  S,  S,  S,  S,  S,  S,  S,  E1},
    /* tRPar    */  { R,  R,  R,  R,  E2, E4, E2, R,  R },
    /* tEndStr  */  { S,  S,  S,  S,  S,  E4, S,  E3, A }
};

/* Group operators by parsing types, and define name string */
struct opDefine oDefine[tMaxOp] = {
   {"+",        tPlusMin }, /* tAdd;  sum */
   {"-",        tPlusMin }, /* tSub;  subtract */
   {"*",        tMultDiv }, /* tMul;  multiply */
   {"/",        tMultDiv }, /* tDiv;  divide */
   {"^",        tExpon   }, /* tPow;  power */
   {"E",        tExpon   }, /* tEE;   exponentiation: x E y = x * 10^y */
   {"-",        tUminus  }, /* tUmi;  unary minus */
   {"fact",     tFunct   }, /* tFact; x! factorial */
   {"perm",     tFunct   }, /* tPerm; perm(n,r): permutations, n objects, r at a time */
   {"cos",      tFunct   }, /* tCos;  cos(x): cosine */
   {"tan",      tFunct   }, /* tTan;  tan(x): tangent */
   {"sin",      tFunct   }, /* tSin;  sin(x): sine */
   {"acos",     tFunct   }, /* tAcos; acos(x): arcosine */
   {"atan",     tFunct   }, /* tAtan; atan(x): arctangent */
   {"asin",     tFunct   }, /* tAsin; asin(x): arcsine */
/*
   {"sinh",     tFunct   }, /o tSinh, sinh(x): hyperbolic sine o/
   {"cosh",     tFunct   }, /o tCosh, cosh(x): hyperbolic cosine o/
   {"tanh",	    tFunct   }, /o tTanh, tanh(x): hyperbolic tangent o/
*/
   {"exp",      tFunct   }, /* tExp;  exp(x): e^x */
   {"log",      tFunct   }, /* tLog;  log(x): log base e */
   {"",         tFunct   }, /* tUser; user-definde functions */
   {",",        tComma   }, /* tComa; comma */ 
   {"(",        tLPar    }, /* tLpr;  left parenthesis */
   {")",        tRPar    }, /* tRpr;  right parenthesis */
   {"",         tEndStr  }  /* tEof;  end of string */
}; 

byte     opr[MAX_OPR];        /* operator stack */
double   val[MAX_VAL];        /* value stack */
int      tokStack[MAX_OPR];   /* token stack (used in recursive 'parse' calls) */
byte     result[MAX_STRLEN];  /* result buffer */
int      oprTop, valTop;      /* top of operator, value stack */
int      tokTop;              /* top of 'tok' stack (used in recursive 'parse' calls) */
int      err_flg;             /* set to TRUE in case of any error */

byte     firsttok = TRUE;     /* true if first token */
tokEnum  prevtok  = tEof;     /* the token before the last taken */
tokEnum  tok      = tEof;     /* token */
double   tokval   = tEof;     /* token value */

struct   UserFunc userfunc[MAX_USERFUNC]; /* To be Loaded with user-defined functions */
int      n_userfunc = 0;     /* Number of user-defined functions */
int      i_userfunc = 0;     /* index to found user function */

long ScanCount = 0;
long TimerCount = 0;

struct   variables var[MAX_VAR];  /* Array of variables */
int      n_var      = 0;          /* Number of variables */

byte     *lasttok;

/******************************************************************************
  Funzioni locali
******************************************************************************/
static int error(const byte *str) ;
double fact(double n) ;
static int reduce(void) ;
static int get_token(byte **expr);
static int validuserfunc(byte *sName, byte *sDefine);
static int add_var(byte *str);
static int setvar_idx(int idx, double val);
static int variable(byte *s);
static int look_up_usr(byte *s);
static int look_up(byte *s);
static int isdelim(byte c);
static int isconstant(byte c);
static int iswhite(byte c);
double exponent(double t, int n);

#ifndef CASIO_PV
/*****************************************************************************/
int main(void) {
 byte Espressione[MAX_STRLEN];
 byte *pEspr;
 byte sName[MAX_OPRSTR];
 byte sDefine[MAX_STRLEN];
 byte sBuffer[MAX_STRLEN];
 byte *b;
 int Err = 0;
 int iUserFunc, i, blen;
 FILE *fp;

 
     /* Questa parte di gestione del file va tutta rifatta sul Casio */
     fp = fopen("userfunc.txt","r");
     while (n_userfunc < MAX_USERFUNC)
     {
	 /* Legge una linea dal file */
	 if(!fgets(sBuffer, MAX_STRLEN, fp)) /* EOF */
	     break;

	 /* Spezza la linea in nome e definizione della funzione */
	 if (!(b = strchr(sBuffer, '\t'))) /* Cerca il TAB */
	     continue;
	 *b = '\0'; blen = strlen(b + 1);
	 strcpy(sName, sBuffer);
	 strcpy(sDefine, b + 1);
	 sDefine[blen - 1] = '\0';

	 /* Valida la nuova funzione da aggiungere alla lista */
	 /* QUESTO DOVREBBE ESSERE FATTO IN FASE DI SCRITTURA NEL FILE DI UNA NUOVA
	    FUNZIONE, NEL QUAL CASO NON SERVIREBBE DURANTE LA LETTURA! */
	 if (!validuserfunc(sName, sDefine))
	     continue;

	 /* Carica la funzione nella lista */
	 strcpy(userfunc[n_userfunc].sName, sName);
	 strcpy(userfunc[n_userfunc].sDefine, sDefine);
	 n_userfunc++;
     }
     fclose(fp);

     if (n_userfunc) printf("Funzioni definite da utente:\n");
     for (i = 0; i < n_userfunc; i++)
     {
	 printf("%2.2d : %s = %s\n",
		i + 1,
		userfunc[i].sName,
		userfunc[i].sDefine);
     }

     printf("\nInserire espressione (q per uscire)\n");

     while(TRUE)
     {
	 gets(Espressione);
	 if (!strcmp(Espressione,"q"))
	 {
	     printf("Ciao\n");
	     return (0);
	 }

	 /* initialize for expression parsing */
	 parseinit();
	 pEspr = Espressione;
	 Err = parse(&pEspr);

	 /* check and format result */
	 Err = printresult();
	 printf("%s\n",result);
     }
}

#else
/********* ONLY FOR CASIO PV ********************************/

void ScanInc(void)
{
    ScanCount++;
}
void TimerInc(void)
{
    TimerCount++;
}


int LoadUserFunc(void)
{
    FILE_BUF	pzlfb;	/* FLASH data buffer (binaryBuffer 3KB)*/
    FILE_INFO	pzlfinf;	/* FLASH data information */
    byte sName[MAX_OPRSTR];
    byte sDefine[MAX_STRLEN];
    byte sBuffer[MAX_STRLEN];
    byte *b;
    int blen, i = 0;
    
    pzlfb.fsb_main_entry_ = 0x30;	/* modalita' MEMO */
    pzlfb.fsb_sub_entry_  = 0x00;	/* gruppo 'Senza Titolo 1' */
    pzlfb.fsb_scrt_info_  = 0x80;	/* Open Mode */
    pzlfinf.fp            = 0xffff;	/* ricerca dall'inizio */
    
    /* Carica i record del gruppo */
    while(LibFileFindNext(&pzlfb,&pzlfinf,0x00))
    {
	LibFileRead(&pzlfb,&pzlfinf);  	/* Read */
	strcpy(sBuffer, pzlfb.fbuf.text.fsb_text_buf_ );
	
	/* Cerca il 'return' che separa nome da definizione */
	if (!(b = strchr(sBuffer, 0x0D))) 
	    continue;
	*b = '\0'; blen = strlen(b + 1);

	strcpy(sName, sBuffer);
	strcpy(sDefine, b + 1);
	sDefine[blen] = '\0';
	
	/* Valida la nuova funzione da aggiungere alla lista */
	/* NOTA: Per avere funzioni utente 'ricorsive' devo definirle nell'ordine
	   corretto (caricando per prima quella che viene inserita in un'altra) */
/*
	if (!validuserfunc(sName, sDefine))
	    continue;
*/	
	/* Carica la funzione nella lista */
	strcpy(userfunc[n_userfunc].sName, sName);
	strcpy(userfunc[n_userfunc].sDefine, sDefine);
	n_userfunc++;
	
	i++;
    }
    
    return (i);
}
#endif

/*****************************************************************************/
void parseinit(void)
{
	 oprTop = tokTop = 0;
	 valTop = -1;
	 return;
}

/*****************************************************************************/
int printresult(void)
{
	if (valTop != 0) return (error("syntax error"));
	if (!err_flg) sprintf(result, "%-.15g", val[valTop]);
	return(0);
}

/*****************************************************************************/
int getvaltop(double *f)
{
	if ((!valTop) && (!err_flg)){
		*f = val[valTop];
		return (0);
	}
	return(1);
}

/*****************************************************************************/
static int validuserfunc(byte *sName, byte *sDefine)
{
byte *pEspr;

    if (isalpha(sName[0]))
	return(0); 

    /* Exit with 'invalid' if name already used */
    if (variable(sName))
	return (0);
    if(look_up(sName))
	return (0);
    if(look_up_usr(sName))
	return (0);

    /* Controllo sintattico (defining the x variable) */
    add_var(USER_X_VAR);
    pEspr = sDefine;
    oprTop = 0; valTop = -1;
    if(parse(&pEspr))
	return (0);
    
    return(1);
}

/*****************************************************************************/
 static int add_var(byte *str){
    
    if (n_var++ < MAX_VAR) {
	strcpy(var[n_var - 1].sName, str);
	return (1);
    }

    return (0); /* Error allocating a new variable */

}

/*****************************************************************************/
int setvar_name(byte *str, double val){
    register i;
    byte *p;
    
    p = str;
    while(*p) { *p = tolower(*p); p++; }
    for(i = 0; i < n_var; i++)
	if(!strcmp(var[i].sName, str))
	{
	    var[i].dValue = val;
	    return (1);
	}

    /* Allocate and set a new variable */
    if (add_var(str))
	if (setvar_idx(n_var - 1, val))
	    return (1);

    return(0); /* Error setting a variable */
}

/*****************************************************************************/
 static int setvar_idx(int idx, double val){

    if(idx < n_var)
    {
	var[idx].dValue = val;
	return (1);
    }

    return (0); /* Error setting a variable */
}

/*****************************************************************************/
int value(byte *str){
byte intero[MAX_VALUE] = "";
byte decimale[MAX_VALUE] = "";
double f;
int i;

   if (str[0] == '.')
       sscanf(str,".%[0-9]", decimale);
   else
       sscanf(str,"%[0-9].%[0-9]", intero, decimale);

   f = atof(decimale);
   for (i = 1; i <= (int)strlen(decimale); i++)
       f /= 10;
   f += atof(intero);
   tok = tVal;
   tokval = f;

   return (1); /* Always return with no error - A validation routine
		  should be added (for example at the sscanf call) */

}

/*****************************************************************************/
int getvar_name(byte *s, double *val)
{
/* Look for the variable by name, and return its value */
register i;
byte *p;
    
    p = s;
    while(*p) { *p = tolower(*p); p++; }
    for(i = 0; i < n_var; i++)
	if(!strcmp(var[i].sName, s)){
	    *val = var[i].dValue;
	    return(1);
	}

    return(0);

}

/*****************************************************************************/
 static int variable(byte *s)
{
/* Look for the variable by name, and return its value in the value stack */
double f;
    
    if(getvar_name(s, &f)){
	    tokval = f;
	    tok = tVal;
	    return(1);
    }

    return(0);

}


/*****************************************************************************/
 static int error(const byte *str) {
        err_flg = TRUE;
	strcpy (result, str);
	return (err_flg);
}

/******************************************************************************
Scan the input string to extract the next token (constant, variable or operator)
Input: pointer to current position in expression string.
       Pointer is updated to the end of the first token found.
Return TRUE in case of error 
******************************************************************************/
 static int get_token(byte **expr){
    char temparray[MAX_OPRSTR];
    char *temp = temparray;
    char *start;

    /* Cut trailing blanks off */
    while(iswhite(**expr)) (*expr)++;

    if (**expr == '\0') 
    { /* End of string */
	tok = tEof;

	prevtok = tok;
	return (0);
    }

    if(isdelim(**expr))
    { /* Look for an operator which is a one-char delimiter as well (+, -, ...) */
	start = temp;
	*temp=**expr;
	temp++;
	(*expr)++;
	*temp = '\0';
	if (!look_up(start))
	    return (error("Undefined delimiter found"));

	/* check for unary minus */
	if (tok == tSub) {
	    if (prevtok != tVal && prevtok != tRpr) {
		tok = tUmi;
	    }
	}
	
	prevtok = tok;
	return (0);
    }

    if(isalpha(**expr)) /* libc.a */
    { /* Scan up to next delimiter */
	start = temp;
	while(!isdelim(**expr))
	{
	    *temp = **expr;
	    temp++;
	    (*expr)++; /* update current position */
	}
	*temp = '\0';
	if(!look_up(start)) /* Look for a pre-loaded operator */
	    if (!look_up_usr(start)) /* Look for a user-defined function */
		if (!variable(start)) /* Operator not found. Evaluate it as a variable  */
		    return(error("Undefined variable found"));

	prevtok = tok;
	return(0);
    }

    if(isconstant(**expr))
    { /* Scan up to next delimiter */
	start = temp;
	while(!isdelim(**expr))
	{
	    *temp = **expr;
	    temp++;
	    (*expr)++;
	}
	*temp = '\0';
	if(!value(start))
	    return(error("Error evaluating constant"));

	prevtok = tok;
	return(0);
    }

    return (error("Tokenization error")); /* Undefined error while tokenizing (syntax) */
 
}

/*****************************************************************************/
 static int look_up_usr(byte *s)
{
    byte *p;
    
    if (n_userfunc)
    {
	p = s;
	while(*p) { *p = tolower(*p); p++; }
	for(i_userfunc = 0; i_userfunc < n_userfunc; i_userfunc++)
	    if(!strcmp(userfunc[i_userfunc].sName, s)){
		tok = tUser;
		return (1);    
	    }
    }

    return (0);
}

/*****************************************************************************/
 static int look_up(byte *s)
{
    register i;
    byte *p;
    
    p = s;
    while(*p) { *p = tolower(*p); p++; }
    for(i = 0; i < tMaxOp; i++)
	if(!strcmp(oDefine[i].oString, s)){
		tok = i;
		return (1);    
	}
    
    return (0);
}

/*****************************************************************************/
 static int isdelim(byte c)
{
	if(strchr(" ,+-<>/*^=()", c) || c == 9 || c == '\n' || c == 0)
		return (1);
	return (0);
}

/*****************************************************************************/ 
 static int isconstant(byte c)
{
    if (isdigit(c) || strchr(".", c)) return (1); 
    else return (0);
}

/*****************************************************************************/
 static int iswhite(byte c)
{
	if(c == ' ' || c == '\t') return (1);
	else return (0);
}

/*****************************************************************************/
 static int shift(byte **str) {
    if (tok == tVal) {
	/* Push a value */
        if (++valTop >= MAX_VAL) 
            return (error("val stack exhausted")); 
        val[valTop] = tokval;
    } else {
	/* Push an operator */
        if (++oprTop >= MAX_OPR - 1)
			return (error("opr stack exhausted")); 

	if (tok == tUser)
	    /* For user-defined functions push its index first */
	    opr[oprTop++] = (byte)i_userfunc;

	opr[oprTop] = (byte)tok;

    }

    if (get_token(str)) return (1); /* Get next token */

    return (0);
}

/*****************************************************************************/
 double fact(double n) {
    double i, t;
    for (t = 1, i = 1; i <= n; i++)
        t *= i;
    return (t);
}

/*****************************************************************************/
 double exponent(double t, int n) {
    int i;

    if (n > 0)
	for (i = 1; i <= n; i++)
	    t *= 10.0;
    else if (n < 0)
	for (i = 1; i <= -n; i++)
	    t /= 10.0;

    return (t);
}

/*****************************************************************************/
 static int reduce(void) {
int i;
byte *pStr;

    switch(opr[oprTop]) {
    case tAdd:
        /* apply E := E + E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] + val[valTop];
        valTop--;
        break;
    case tSub:
        /* apply E := E - E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] - val[valTop];
        valTop--;
        break;
    case tMul:
        /* apply E := E * E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] * val[valTop];
        valTop--;
        break;
    case tDiv:
        /* apply E := E / E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] / val[valTop];
        valTop--;
        break;
    case tUmi:
        /* apply E := -E */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = -val[valTop];
        break;
    case tPow:
        /* apply E := E ^ E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = pow(val[valTop-1], val[valTop]) ; 
        valTop--;
        break;
    case tEE:
        /* apply E := E * 10^ E */
        if (valTop < 1)  return (error("syntax error"));
        val[valTop-1] = exponent(val[valTop-1], (int)val[valTop]);
        valTop--;
        break;
    case tFact:
        /* apply E := f(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = fact(val[valTop]);
        break;
    case tSin:
        /* apply E := sin(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = sin(val[valTop]);
        break;
    case tCos:
        /* apply E := cos(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = cos(val[valTop]);
        break;
    case tTan:
        /* apply E := tan(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = tan(val[valTop]);
        break;
    case tAcos:
        /* apply E := acos(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = acos(val[valTop]);
        break;
    case tAtan:
        /* apply E := atan(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = atan(val[valTop]);
        break;
    case tAsin:
        /* apply E := asin(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = asin(val[valTop]);
        break;
/*
    case tSinh:
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = sinh(val[valTop]);
        break;
    case tCosh:
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = cosh(val[valTop]);
        break;
    case tTanh:
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = tanh(val[valTop]);
        break;
*/
    case tExp:
        /* apply E := e ^ E */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = exp(val[valTop]);
        break;
    case tLog:
        /* apply E := ln E (base e) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = log(val[valTop]);
        break;
/*
    case tComb:
        /o apply E := c(N,R) o/
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = fact(val[valTop-1])/
            (fact(val[valTop]) * fact(val[valTop-1]-val[valTop]));
        valTop--;
        break;
*/
    case tUser:
	/* Evaluate the user-defined function */
	/* The top stack value will set the 'x' variable.
	   This name is reserved to user-function definitions.
	   Example: if we defined two user functions: f = 2*x, g = x^2,
	   and we are evaluating the expression f(2+g(9)),
	   when whe reach the reduction of 'g', the x is set to 9.
	   Then 2+g(9)=83 will be pushed to the value stack, and when
	   reducing 'f' the x will be set to 83.
	   Note recurrency in calling parse() */

	/* Set the 'x' variable value to top stack value */
	if(!setvar_name(USER_X_VAR, val[valTop--]))
	     return (error("Error while setting X in user function"));

	/* Save 'tok' value before parsing the user function */
	if (++tokTop >= MAX_VAL) 
            return (error("tok stack exhausted")); 
	tokStack[tokTop] = tok;
	
	/* Parse the user-defined expression
	   leaving its result in the top stack value */
	oprTop--;
	pStr = userfunc[opr[oprTop]].sDefine;
	if(parse(&pStr))
	    return (error("Error parsing user defined function"));

	/* Retrieve the saved 'tok' */
	tok = tokStack[tokTop--];
	
	break;

    case tPerm:
        /* apply E := p(N,R) */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = fact(val[valTop-1])/fact(val[valTop-1]-val[valTop]);
        valTop--;
        break;
    case tRpr:
        /* pop () off stack */
        oprTop--;
        break;
    }
    oprTop--;
    return (0);
}

/*****************************************************************************/
int parse(byte **str){

    opr[oprTop] = tEof;
    firsttok = TRUE;
    err_flg = FALSE;

    if (firsttok) { /* Alla prima scansione, inizializza le variabili */
        firsttok = FALSE;
        prevtok = tEof;
    }

    if (get_token(str)) /* prende il primo token */
	return (1);

    while(1) {

        /* input is value */
        if (tok == tVal) { 
            /* shift token to value stack */
            if (shift(str)) return (1);
            continue;
        }

        /* input is operator */
        switch(parsetable
	       [oDefine[opr[oprTop]].oType]
	       [oDefine[tok].oType])
	{
        case R:
            if (reduce()) return (1);
            break;
        case S:
            if (shift(str)) return (1);
            break;
        case A:
            /* accept final result */
	    if (oprTop > 0) /* It means this is a recursive call */ 
		tok = opr[oprTop];
            return (0);
        case E1:
            return (error("missing right parenthesis"));
        case E2:
            return (error("missing operator"));
        case E3:
            return (error("unbalanced right parenthesis"));
        case E4:
            return (error("invalid function argument"));
        }
    }
}

