/******************************************************************************
  Include di sistema
******************************************************************************/
#include <stdrom.h>
#include <dummy.h> /* Used to compile with math functions */

/******************************************************************************
  Include applicazione
******************************************************************************/
#include "parser.h"

/******************************************************************************
  Variabili globali
******************************************************************************/
/* Tabella di parsing */
char parseTbl[tMaxOp][tMaxOp] = {
 /* stk        --------------input ------------------------------------------ */
 /*            +  -  *  /  ^  E  M  f  p  c  t  s  ac at as e  l  ,  (  )  $  */
 /*            -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- */
 /* +    */  { R, R, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, R, S, R, R },
 /* -    */  { R, R, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, R, S, R, R },
 /* *    */  { R, R, R, R, S, S, S, S, S, S, S, S, S, S, S, S, S, R, S, R, R },
 /* /    */  { R, R, R, R, S, S, S, S, S, S, S, S, S, S, S, S, S, R, S, R, R },
 /* ^    */  { R, R, R, R, S, S, S, S, S, S, S, S, S, S, S, S, S, R, S, R, R },
 /* E    */  { R, R, R, R, S, S, S, S, S, S, S, S, S, S, S, S, S, R, S, R, R },
 /* M    */  { R, R, R, R, R, R, S, S, S, S, S, S, S, S, S, S, S, R, S, R, R },
 /* fact */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* perm */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* cos  */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* tan  */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* sin  */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* acos */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* atan */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* asin */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* exp  */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* log  */  { R, R, R, R, S, S, E4,S, S, S, S, S, S, S, S, S, S, E4,S, R, R },
 /* ,    */  { R, R, R, R, R, R, R, R, R, R, R, R, R, R, R, R, R, E4,R, R, E4},
 /* (    */  { S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, E1},
 /* )    */  { R, R, R, R, R, R, R, E3,E3,E3,E3,E4,E4,E4,E4,E4,E4,E4,E2,R, R },
 /* $    */  { S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, E4,S, E3,A }
};

char parsetable[tMaxType][tMaxType] = {
    /*                ------------- INPUT -------------- */
    /*                +   *   ^   M   f   ,   (   )   $  */
    /*  STACK         --  --  --  --  --  --  --  --  -- */
    /* tPlusMin */  { R,  S,  S,  S,  S,  R,  S,  R,  R },
    /* tMultDiv */  { R,  R,  S,  S,  S,  R,  S,  R,  R },
    /* tExpon   */  { R,  R,  S,  S,  S,  R,  S,  R,  R },
    /* tUminus  */  { R,  R,  R,  S,  S,  R,  S,  R,  R },
    /* tFunct   */  { R,  R,  S,  E4, S,  E4, S,  R,  R },
    /* tComma   */  { R,  R,  R,  R,  R,  E4, R,  R,  E4},
    /* tLPar    */  { S,  S,  S,  S,  S,  S,  S,  S,  E1},
    /* tRPar    */  { R,  R,  R,  R,  E2, E4, E2, R,  R },
    /* tEndStr  */  { S,  S,  S,  S,  S,  E4, S,  E3, A }
};

/* An array to contain operator names */
/* NB : having two operators with names like 'cos' and 'cosh' is not allowed;
        'cos' and 'hcos' it's ok instead (because of the scanning
        procedure used to search tokens: function nexttok) */
char tokName [tMaxOp][MAX_OPRSTR] = {
   "+",        /* tAdd,  sum */
   "-",        /* tSub, subtract */
   "*",        /* tMul, multiply */
   "/",        /* tDiv, divide */
   "^",        /* tPow, power */
   "E",        /* tEE, exponentiation: x E y = x * 10^y */
   "-",        /* tUmi, unary minus */
   "fact",     /* tFact, x! factorial */
   "perm",     /* tPerm, perm(n,r): permutations, n objects, r at a time */
   "cos",      /* tCos, cos(x): cosine */
   "tan",      /* tTan, tan(x): tangent */
   "sin",      /* tSin, sin(x): sine */
   "acos",     /* tAcos, acos(x): arcosine */
   "atan",     /* tAtan, atan(x): arctangent */
   "asin",     /* tAsin, asin(x): arcsine */
   "exp",      /* tExp, exp(x): e^x */
   "log",      /* tLog, log(x): log base e */
/* "comb",     /o tComb, comb(n, r): combinations, n objects, r at a time */
   ",",        /* tComa, comma */ 
   "(",        /* tLpr, left parenthesis */
   ")",        /* tRpr, left parenthesis */
   ""          /* tEof, End of string */
};

/* Group operators by parsing types, and give them a string */
struct opDefine oDefine[tMaxOp] = {
   {"+",        tPlusMin }, /* tAdd;  sum */
   {"-",        tPlusMin }, /* tSub;  subtract */
   {"*",        tMultDiv }, /* tMul;  multiply */
   {"/",        tMultDiv }, /* tDiv;  divide */
   {"^",        tExpon   }, /* tPow;  power */
   {"E",        tExpon   }, /* tEE;   exponentiation: x E y = x * 10^y */
   {"-",        tUminus  }, /* tUmi;  unary minus */
   {"fact",     tFunct   }, /* tFact; x! factorial */
   {"perm",     tFunct   }, /* tPerm; perm(n,r): permutations, n objects, r at a time */
   {"cos",      tFunct   }, /* tCos;  cos(x): cosine */
   {"tan",      tFunct   }, /* tTan;  tan(x): tangent */
   {"sin",      tFunct   }, /* tSin;  sin(x): sine */
   {"acos",     tFunct   }, /* tAcos; acos(x): arcosine */
   {"atan",     tFunct   }, /* tAtan; atan(x): arctangent */
   {"asin",     tFunct   }, /* tAsin; asin(x): arcsine */
   {"exp",      tFunct   }, /* tExp;  exp(x): e^x */
   {"log",      tFunct   }, /* tLog;  log(x): log base e */
   {"",         tFunct   }, /* tUser; user-definde functions */
   {",",        tComma   }, /* tComa; comma */ 
   {"(",        tLPar    }, /* tLpr;  left parenthesis */
   {")",        tRPar    }, /* tRpr;  right parenthesis */
   {"",         tEndStr  }  /* tEof;  end of string */
}; 

char     opr[MAX_OPR];        /* operator stack */
double   val[MAX_VAL];        /* value stack */
int      tokStack[MAX_OPR];   /* token stack (used in recursive 'parse' calls) */
char     result[MAX_STRLEN];  /* result buffer */
int      oprTop, valTop;      /* top of operator, value stack */
int      tokTop;              /* top of 'tok' stack (used in recursive 'parse' calls) */
int      err_flg;             /* TRUE in case of ay error */

char     firsttok = TRUE;     /* true if first token */
tokEnum  prevtok  = tEof;     /* the token before the last taken */
tokEnum  tok      = tEof;     /* token */
double   tokval   = tEof;     /* token value */

struct   UserFunc userfunc[MAX_USERFUNC]; /* To be Loaded with user-defined functions */
int      n_userfunc = 0;     /* Number of user-defined functions */
int      i_userfunc = 0;     /* index to found user function */

struct   variables var[MAX_VAR];  /* Array of variables */
int      n_var      = 0;          /* Number of variables */

char     *lasttok;

/******************************************************************************
  Funzioni locali
******************************************************************************/
static int error(const char *str) ;
static double fact(double n) ;
static int reduce(void) ;
static int nexttok(char * str);
static double value(char *str);
static double exponent(double t, int n);

/*****************************************************************************
int main(void) {
char Espressione[MAX_STRLEN];
int Err = 0;

     printf("Inserire espressione (q per uscire)\n");

     while(TRUE)
     {
	 gets(Espressione);
	 if (!strcmp(Espressione,"q"))
	 {
	     printf("Ciao\n");
	     return (0);
	 }

	 Err = parse(Espressione);
	 printf("%s\n",result);
     }
}
*/

/*****************************************************************************/
static double value(char *str){
char intero[MAX_VALUE] = "";
char decimale[MAX_VALUE] = "";
char appoggio[MAX_STRLEN + 1] = "";
double f;
int i;

   if (str[0] == '.'){
       strcpy(appoggio, "0");
       strcat(appoggio, str);
   }
   else
       strcpy(appoggio, str);

   sscanf(appoggio,"%[0-9].%[0-9]", intero, decimale);

   f = atof(decimale);
   for (i = 1; i <= (int)strlen(decimale); i++)
       f /= 10;
   f += atof(intero);
   return (f);

}

/*****************************************************************************/
static int error(const char *str) {
	strcpy (result, str);
	return (1);
}

/*****************************************************************************/
static int nexttok(char * str){
int i;
char *s;
char *found = NULL;

/* DA FARE ANCORA:
- Distinguere operatori con parte iniziale del nome uguale (tipo cos e cosh)
*/

    if (firsttok) { /* Alla prima scansione, inizializza le variabili */
        firsttok = FALSE;
        prevtok = tEof;
	lasttok = str; /* posiziona il puntatore della scansione all'inizio */
    }

    /* salta eventuali spazi o altro da ignorare in testa alla stringa */
    while (lasttok[0] == ' ' ||
	   lasttok[0] == '\t' ||
	   lasttok[0] == '\n')
	lasttok++;

    /* Cerca tutti gli operatori presenti nella stringa e memorizza il primo */ 
    for (i = 0; i < tMaxOp - 1; i++)
	if (s = strstr(lasttok, tokName[i]))
	    if ((s < found) || !found)
	    {
		found = s;
		tok = i;
	    }

    if (found) /* trovato l'operatore */
	if (found != lasttok) /* ma c' un valore prima */
	{
	    tok = tVal;
            tokval = value(lasttok);
	    /* tokval = atof(lasttok); */
	    lasttok = found;
	}
	else
        {
	    /* check for unary minus */
	    if (tok == tSub) {
		if (prevtok != tVal && prevtok != tRpr) {
			tok = tUmi;
		}
	    }
	    lasttok += (int)strlen(tokName[tok]);
	}
    else
	/* Non ha trovato alcun operatore ( un valore o fine stringa) */
	if (strlen(lasttok) > 0)
	{
	    tok = tVal; /*  un valore */	    
            tokval = value(lasttok);
	    /* tokval = atof(lasttok); */
	    lasttok += (int)strlen(lasttok);
	}
	else
	    tok = tEof; 

    prevtok = tok;
    return (0);
}


/*****************************************************************************/
static int shift(char * str) {
    if (tok == tVal) {
        if (++valTop >= MAX_VAL) 
            return (error("val stack exhausted")); 
        val[valTop] = tokval;
    } else {
        if (++oprTop >= MAX_OPR)
			return (error("opr stack exhausted")); 
        opr[oprTop] = (char)tok;
    }
    if (nexttok(str)) return (1);
    return (0);
}

/*****************************************************************************/
static double fact(double n) {
    double i, t;
    for (t = 1, i = 1; i <= n; i++)
        t *= i;
    return (t);
}

/*****************************************************************************/
static double exponent(double t, int n) {
    int i;

    if (n > 0)
	for (i = 1; i <= n; i++)
	    t *= 10.0;
    else if (n < 0)
	for (i = 1; i <= -n; i++)
	    t /= 10.0;

    return (t);
}

/*****************************************************************************/
static int reduce(void) {
    switch(opr[oprTop]) {
    case tAdd:
        /* apply E := E + E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] + val[valTop];
        valTop--;
        break;
    case tSub:
        /* apply E := E - E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] - val[valTop];
        valTop--;
        break;
    case tMul:
        /* apply E := E * E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] * val[valTop];
        valTop--;
        break;
    case tDiv:
        /* apply E := E / E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = val[valTop-1] / val[valTop];
        valTop--;
        break;
    case tUmi:
        /* apply E := -E */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = -val[valTop];
        break;
    case tPow:
        /* apply E := E ^ E */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = pow(val[valTop-1], val[valTop]) ; 
        valTop--;
        break;
    case tEE:
        /* apply E := E * 10^ E */
        if (valTop < 1)  return (error("syntax error"));
		if (val[valTop] != (int)val[valTop])
				return (error("invalid exponent"));
        val[valTop-1] = exponent(val[valTop-1], (int)val[valTop]);
        valTop--;
        break;
    case tFact:
        /* apply E := f(E) */
        if (valTop < 0) return (error("syntax error"));
		if (val[valTop] != (int)val[valTop])
				return (error("fact invalid argument"));
        val[valTop] = fact(val[valTop]);
        break;
    case tSin:
        /* apply E := sin(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = sin(val[valTop]);
        break;
    case tCos:
        /* apply E := cos(E) */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = cos(val[valTop]);
        break;
    case tTan:
        /* apply E := tan(E) */
        if (valTop < 0) return (error("syntax error"));
		if (fabs(val[valTop] - M_PI_2) < MIN_PRECISION)
			return (error("tan invalid argument"));
        val[valTop] = tan(val[valTop]);
        break;
	case tAcos:
        /* apply E := acos(E) */
        if (valTop < 0) return (error("syntax error"));
		if (val[valTop] < -1.0 || val[valTop] > 1.0)
			return (error("acos invalid argument"));
        val[valTop] = acos(val[valTop]);
        break;
	case tAtan:
        /* apply E := atan(E) */
        if (valTop < 0) return (error("syntax error"));
		if (val[valTop] < -1.0 || val[valTop] > 1.0)
			return (error("atan invalid argument"));
        val[valTop] = atan(val[valTop]);
        break;
	case tAsin:
        /* apply E := asin(E) */
        if (valTop < 0) return (error("syntax error"));
		if (val[valTop] < -1.0 || val[valTop] > 1.0)
			return (error("asin invalid argument"));
        val[valTop] = asin(val[valTop]);
        break;
	case tExp:
        /* apply E := e ^ E */
        if (valTop < 0) return (error("syntax error"));
        val[valTop] = exp(val[valTop]);
        break;
	case tLog:
        /* apply E := ln E (base e) */
        if (valTop < 0) return (error("syntax error"));
		if (val[valTop]) return (error("log invalid argument"));
        val[valTop] = log(val[valTop]);
        break;
	/*
    case tComb:
        /o apply E := c(N,R) o/
        if (valTop < 1) return error("syntax error");
        val[valTop-1] = fact(val[valTop-1])/
            (fact(val[valTop]) * fact(val[valTop-1]-val[valTop]));
        valTop--;
        break;
*/
    case tPerm:
        /* apply E := p(N,R) */
        if (valTop < 1) return (error("syntax error"));
        val[valTop-1] = fact(val[valTop-1])/fact(val[valTop-1]-val[valTop]);
        valTop--;
        break;
    case tRpr:
        /* pop () off stack */
        oprTop--;
        break;
    }
    oprTop--;
    return (0);
}

/*****************************************************************************/
void parseinit(void)
{
	 oprTop = tokTop = 0;
	 valTop = -1;
	 return;
}

/*****************************************************************************/
int printresult(void)
{
	if (valTop != 0) return (error("syntax error"));
	if (!err_flg) sprintf(result, "%-.15g", val[valTop]);
	return(0);
}

/*****************************************************************************/
int parse(char *str){

    /* initialize for next expression */
    oprTop = 0; valTop = -1;
    opr[oprTop] = tEof;
    firsttok = TRUE;
    if (nexttok(str)) /* prende il prossimo token */
	return (1);

    while(1) {

        /* input is value */
        if (tok == tVal) { 
            /* shift token to value stack */
            if (shift(str)) return (1);
            continue;
        }

        /* input is operator */
        switch(parseTbl[opr[oprTop]][tok]) {
        case R:
            if (reduce()) return (1);
            break;
        case S:
            if (shift(str)) return (1);
            break;
        case A:
            /* accept and format final result */
            if (valTop != 0) return (error("syntax error"));
            sprintf(result, "%-.15g", val[valTop]);
            return (0);
        case E1:
            return (error("missing right parenthesis"));
        case E2:
            return (error("missing operator"));
        case E3:
            return (error("unbalanced right parenthesis"));
        case E4:
            return (error("invalid function argument"));
        }
    }
}

