#ifndef PARSER_H_DEF
#define PARSER_H_DEF

/******************************************************************************
  Costanti
******************************************************************************/
#define TRUE            1
#define FALSE           0

#define MAX_OPR         50
#define MAX_VAL         50
#define MAX_VAR         20
#define MAX_STRLEN      250
#define MAX_VALUE       40
#define MAX_OPRSTR      20
#define MAX_USERFUNC    20
#define USER_X_VAR      "x" /* reserved to the user-defined functions variable */

#define MIN_PRECISION   1.0E-20
#define MAX_PRECISION   1.0E+20
#define M_E         2.71828182845904523536028747135266250
#define M_LOG2E     1.44269504088896340735992468100189214
#define M_LOG10E    0.434294481903251827651128918916605082
#define M_LN2       0.693147180559945309417232121458176568
#define M_LN10      2.30258509299404568401799145468436421
#define M_PI        3.14159265358979323846264338327950288
#define M_PI_2      1.57079632679489661923132169163975144
#define M_PI_4      0.785398163397448309615660845819875721
#define M_1_PI      0.318309886183790671537767526745028724
#define M_2_PI      0.636619772367581343075535053490057448
#define M_2_SQRTPI  1.12837916709551257389615890312154517
#define M_SQRT2     1.41421356237309504880168872420969808
#define M_SQRT1_2   0.707106781186547524400844362104849039

/******************************************************************************
  Tipi
******************************************************************************/

/* actions */
typedef enum {
    S,                 /* shift */
    R,                 /* reduce */
    A,                 /* accept */
    E1,                /* error: missing right parenthesis */
    E2,                /* error: missing operator */
    E3,                /* error: unbalanced right parenthesis */
    E4                 /* error: invalid function argument */
} actEnum;

/* tokens types (for the parsing table) */
typedef enum {
	tPlusMin,
	tMultDiv,
	tExpon,
	tUminus,
	tFunct,
	tComma,
	tLPar,
	tRPar,
	tEndStr,
	tMaxType
} tokType;

/* tokens */
typedef enum {
	tAdd,               /* + */
	tSub,               /* - */
	tMul,               /* * */
	tDiv,               /* / */
	tPow,               /* ^ (power) */
	tEE,                /* exponentiation: x E y = x * 10^y */
	tUmi,               /* - (unary minus) */
	tFact,              /* f(x): factorial */
	tPerm,              /* p(n,r): permutations, n objects, r at a time */
	tCos,               /* cos(x): cosine */
	tTan,               /* tan(x): tangent */
	tSin,               /* sin(x): sine */
	tAcos,              /* acos(x): arcosine */
	tAtan,              /* atan(x): arctangent */
	tAsin,              /* asin(x): arcsine */
/*
      tSinh,              /o sinh(x): hyperbolic sine o/
      tCosh,              /o cosh(x): hyperbolic cosine o/
      tTanh,              /o tanh(x): hyperbolic tangent o/
*/
	tExp,               /* exp(x): e^x */
	tLog,               /* log(x): log base e */
	/* tComb,           /o c(n,r): combinations, n objects, r at a time */
	tUser,              /* User-defined function */
	tComa,              /* comma */
	tLpr,               /* ( */
	tRpr,               /* ) */
	tEof,               /* end of string */     
	tMaxOp,             /* maximum number of operators */
	/* non-operators */
	tVal                /* value */
} tokEnum;

/* Operator Token structure */
typedef struct opDefine
{
    byte    oString[MAX_OPRSTR];
    tokType oType;
};

typedef struct UserFunc
{
    byte   sName[MAX_OPRSTR];
    byte   sDefine[MAX_STRLEN];
};

typedef struct variables
{
    byte   sName[MAX_OPRSTR];
    double dValue;
};

/* FUNZIONI DA AGGIUNGERE:
double sqrt(double);
double fabs(double);
double floor(double);
double ceil(double);
double nearest (double);
double fmod(double, double);

ALTRO DA FARE:
 - rendere le funzioni insensibili a maiuscole/minuscole

 - migliorare il tokenizer (soprattuto per le costanti numeriche)

 - generalizzare la tabella di parsing
    definendo categorie di funzioni che si comportano nella stessa maniera per le priorit 
    e dichiarando ogni operatore con un ordinale di categoria (da cui ricavare l'azione
    associata nella tabella di parsing) e un ordinale di funzione (con cui accedere al
    ramo opportuno dentro la funzione reduce()). A questo punto nella stessa struttura 
    potrebbe andare anche la stringa con il nome dell'operatore ("cos", ecc.).

 - funzioni definite dall'utente
	le funzione verrebbe salvata con un suo nome e associata alla stringa che la definisce,
	in cui la variabile sar sostituita con l'espressione che la contiene.
	E' necessario prima generalizzare la tabella di parsing.

    Esempio:
    f(X)  stata definita con la stringa "sin(X)^2 + cos(X)^2" e salvata. 
    Per calcolare l'espressione "f(12 - 2)" ho due possibilit:
    1) espandere l'espressione completa in una stringa: "sin(12 - 2)^2 + cos(12 - 2)^2", che 
       per potrebbe avere dimensioni molto grandi se la f fosse complessa e chiamata
       pi volte ricorsivamente.
    2) al momento in cui incontro l'operatore f e lo devo ridurre (nella funzione reduce()),
       richiamo la parse() su una stringa costruita con la definizione di f(X) e al posto
       della X il top dello stack dei valori. Sarebbe possibile definire anche funzioni
       a pi variabili o con nomi di varibili definite dall'utente.
       Attenzione alla ricorsivit nella chiamata a parse() e al fatto che ora il risultato
       viene sempre messo in un'unica variabile globale, mentre dovrebbe essere messo
       in uno stack di risultati: parse() farebbe il 'push', reduce() farebbe il 'pop'.

 - aggungere operatori e funzioni logiche
    a = b : restituiscono 1 o 0
    a > b
    a >= b
    ...
    if(a,b,c): se a  diverso da 0 restituisce b, altrimenti c 

 - altre funzioni:
    min(a, b)
    max(a, b)
    average(a, b)
    ...

 - costanti
    se si incontrano alcuni token prefissati (PI, e, ecc.) va sostituito
    il valore corrispondente. La stessa cosa si potrebbe realizzare con funzioni
    definite dall'utente ma con prestazioni peggiori.

*/

/******************************************************************************
  Funzioni esportate
******************************************************************************/
int parse(byte **stringa);
void parseinit(void);
int printresult(void);
int LoadUserFunc(void);
int setvar_name(byte *str, double val);
int getvar_name(byte *s, double *val);
int getvaltop(double *f);
int value(byte *str);


#endif

